from __future__ import annotations

import getpass
import inspect
import re
import sys
from collections.abc import Sequence
from dataclasses import dataclass

from .common import (
    IS_SYSTEMD,
    Command,
    OnCalendar,
    When,
    wrap,
)

OnFailureAction = str

UnitName = str


@dataclass
class Job:
    when: When | None
    command: Command
    unit_name: UnitName
    on_failure: Sequence[OnFailureAction]
    kwargs: dict[str, str]


# staticmethod isn't callable directly prior to 3.10
def _email(to: str) -> str:
    return f'{sys.executable} -m dron.notify.email --job %n --to {to}'


class notify:
    @staticmethod
    def email(to: str) -> str:
        return _email(to)

    email_local = _email(to='%u' if IS_SYSTEMD else getpass.getuser())

    # TODO adapt to macos
    desktop_notification = f'{sys.executable} -m dron.notify.ntfy_desktop --job %n'

    telegram = f'{sys.executable} -m dron.notify.telegram --job %n'


def job(
    when: When | None,
    command: Command,
    *,
    unit_name: str | None = None,
    on_failure: Sequence[OnFailureAction] = (notify.email_local,),
    **kwargs,
) -> Job:
    """
    when: if None, then timer won't be created (still allows running job manually)
    unit_name: if None, then will attempt to guess from source code (experimental!)
    """
    assert 'extra_email' not in kwargs, unit_name  # deprecated

    stacklevel: int = kwargs.pop('stacklevel', 1)

    def guess_name() -> str | Exception:
        stack = inspect.stack()
        frame = stack[stacklevel + 1]  # +1 for guess_name itself
        code_context_lines = frame.code_context
        # python should alway keep single line for code context? but just in case
        if code_context_lines is None or len(code_context_lines) != 1:
            return RuntimeError(f"Expected single code context line, got {code_context_lines=}")
        [code_context] = code_context_lines
        code_context = code_context.strip()
        rgx = r'(\w+)\s+='
        m = re.match(rgx, code_context)  # find assignment to variable
        if m is None:
            return RuntimeError(f"Couldn't guess from {code_context=} (regex {rgx=})")
        return m.group(1)

    if unit_name is None:
        guessed_name = guess_name()

        if isinstance(guessed_name, Exception):
            raise RuntimeError(f"{when} {command}: couldn't guess job name: {guessed_name}")

        unit_name = guessed_name

    return Job(
        when=when,
        command=command,
        unit_name=unit_name,
        on_failure=on_failure,
        kwargs=kwargs,
    )


__all__ = (
    'Command',
    'Job',  # todo maybe don't expose it?
    'OnCalendar',
    'OnFailureAction',
    'When',
    'job',
    'notify',
    'wrap',
)



from __future__ import annotations

import importlib.util
import sys
from collections import OrderedDict
from collections.abc import Iterable, Iterator
from concurrent.futures import ProcessPoolExecutor
from difflib import unified_diff
from itertools import tee
from pathlib import Path
from subprocess import check_call
from typing import NamedTuple

import click

from . import launchd, systemd
from .api import Job, UnitName
from .common import (
    ALWAYS,
    IS_SYSTEMD,
    Body,
    State,
    Unit,
    UnitFile,
    UnitState,
    logger,
    unwrap,
)
from .systemd import _systemctl

# todo appdirs?
DRON_DIR = Path('~/.config/dron').expanduser()
DRON_UNITS_DIR = DRON_DIR / 'units'
DRON_UNITS_DIR.mkdir(parents=True, exist_ok=True)


def verify_units(pre_units: list[tuple[UnitName, Body]]) -> None:
    # need an inline import here in case we modify this variable from cli/tests
    from .common import VERIFY_UNITS

    if not VERIFY_UNITS:
        return

    if len(pre_units) == 0:
        # otherwise systemd analayser would complain if we pass zero units
        return

    if not IS_SYSTEMD:
        for unit_name, body in pre_units:
            launchd.verify_unit(unit_name=unit_name, body=body)
    else:
        systemd.verify_units(pre_units=pre_units)


def verify_unit(*, unit_name: UnitName, body: Body) -> None:
    return verify_units([(unit_name, body)])


def write_unit(*, unit: Unit, body: Body, prefix: Path = DRON_UNITS_DIR) -> None:
    unit_file = prefix / unit

    logger.info(f'writing unit file: {unit_file}')
    verify_unit(unit_name=unit_file.name, body=body)
    unit_file.write_text(body)


def _daemon_reload() -> None:
    if IS_SYSTEMD:
        check_call(_systemctl('daemon-reload'))
    else:
        # no-op under launchd
        pass


def managed_units(*, with_body: bool) -> State:
    if IS_SYSTEMD:
        yield from systemd.systemd_state(with_body=with_body)
    else:
        yield from launchd.launchd_state(with_body=with_body)


def make_state(jobs: Iterable[Job]) -> State:
    pre_units = []
    names: set[Unit] = set()
    for j in jobs:
        uname = j.unit_name

        assert uname not in names, j
        names.add(uname)

        if IS_SYSTEMD:
            s = systemd.service(unit_name=uname, command=j.command, on_failure=j.on_failure, **j.kwargs)
            pre_units.append((uname + '.service', s))

            when = j.when
            if when is None:
                # manual job?
                continue
            if when == ALWAYS:
                continue
            t = systemd.timer(unit_name=uname, when=when)
            pre_units.append((uname + '.timer', t))
        else:
            p = launchd.plist(unit_name=uname, command=j.command, on_failure=j.on_failure, when=j.when)
            pre_units.append((uname + '.plist', p))

    verify_units(pre_units)

    for unit_file, body in pre_units:
        yield UnitState(
            unit_file=DRON_UNITS_DIR / unit_file,
            body=body,
            cmdline=None,  # ugh, a bit crap, but from this code path cmdline doesn't matter
        )


# TODO bleh. too verbose..
class Update(NamedTuple):
    unit_file: UnitFile
    old_body: Body
    new_body: Body

    @property
    def unit(self) -> str:
        return self.unit_file.name


class Delete(NamedTuple):
    unit_file: UnitFile

    @property
    def unit(self) -> str:
        return self.unit_file.name


class Add(NamedTuple):
    unit_file: UnitFile
    body: Body

    @property
    def unit(self) -> str:
        return self.unit_file.name


Action = Update | Delete | Add
Plan = Iterable[Action]

# TODO ugh. not sure how to verify them?


def compute_plan(*, current: State, pending: State) -> Plan:
    # eh, I feel like i'm reinventing something already existing here...
    currentd = OrderedDict((x.unit_file, unwrap(x.body)) for x in current)
    pendingd = OrderedDict((x.unit_file, unwrap(x.body)) for x in pending)

    units = [c for c in currentd if c not in pendingd] + list(pendingd.keys())
    for u in units:
        in_cur = u in currentd
        in_pen = u in pendingd
        if in_cur:
            if in_pen:
                # TODO not even sure I should emit it if bodies match??
                yield Update(unit_file=u, old_body=currentd[u], new_body=pendingd[u])
            else:
                yield Delete(unit_file=u)
        else:
            if in_pen:
                yield Add(unit_file=u, body=pendingd[u])
            else:
                raise AssertionError("Can't happen")


# TODO it's not apply, more like 'compute' and also plan is more like a diff between states?
def apply_state(pending: State) -> None:
    current = list(managed_units(with_body=True))

    pending_units = {s.unit_file.name for s in pending}

    def is_always_running(unit_path: Path) -> bool:
        name = unit_path.stem
        has_timer = f'{name}.timer' in pending_units
        # TODO meh. not ideal
        return not has_timer

    plan = list(compute_plan(current=current, pending=pending))

    deletes: list[Delete] = []
    adds: list[Add] = []
    _updates: list[Update] = []

    for a in plan:
        if isinstance(a, Delete):
            deletes.append(a)
        elif isinstance(a, Add):
            adds.append(a)
        elif isinstance(a, Update):
            _updates.append(a)
        else:
            raise TypeError("Can't happen", a)

    if len(deletes) == len(current) and len(deletes) > 0:
        msg = "Trying to delete all managed jobs"
        if click.confirm(f'{msg}. Are you sure?', default=False):
            pass
        else:
            raise RuntimeError(msg)

    Diff = list[str]
    nochange: list[Update] = []
    updates: list[tuple[Update, Diff]] = []

    for u in _updates:
        unit = a.unit
        diff: Diff = list(
            unified_diff(
                u.old_body.splitlines(keepends=True),
                u.new_body.splitlines(keepends=True),
            )
        )
        if len(diff) == 0:
            nochange.append(u)
        else:
            updates.append((u, diff))

    # TODO list unit names here?
    logger.info(f'no change: {len(nochange)}')
    logger.info(f'disabling: {len(deletes)}')
    logger.info(f'updating : {len(updates)}')
    logger.info(f'adding   : {len(adds)}')

    for a in deletes:
        if IS_SYSTEMD:
            # TODO stop timer first?
            check_call(_systemctl('stop', a.unit))
            check_call(_systemctl('disable', a.unit))
        else:
            launchd.launchctl_unload(unit=Path(a.unit).stem)
    for a in deletes:
        (DRON_UNITS_DIR / a.unit).unlink()

    for u, diff in updates:
        unit = u.unit
        unit_file = u.unit_file
        logger.info(f'updating {unit}')
        for d in diff:
            sys.stderr.write(d)
        write_unit(unit=u.unit, body=u.new_body)
        if IS_SYSTEMD:
            if unit.endswith('.service') and is_always_running(unit_file):
                # persistent unit needs a restart to pick up change
                _daemon_reload()
                check_call(_systemctl('restart', unit))
        else:
            launchd.launchctl_reload(unit=Path(unit).stem, unit_file=unit_file)

        if unit.endswith('.timer'):
            _daemon_reload()
            # NOTE: need to be careful -- seems that job might trigger straightaway if it's on interval schedule
            # so if we change something unrelated (e.g. whitespace), it will start all jobs at the same time??
            check_call(_systemctl('restart', u.unit))

    for a in adds:
        logger.info(f'adding {a.unit_file}')
        # TODO when we add, assert that previous unit wasn't managed? otherwise we overwrite something
        write_unit(unit=a.unit, body=a.body)

    # need to load units before starting the timers..
    _daemon_reload()

    for a in adds:
        unit_file = a.unit_file
        unit = unit_file.name
        logger.info(f'enabling {unit}')
        if unit.endswith('.service'):
            # quiet here because it warns that "The unit files have no installation config"
            # TODO maybe add [Install] section? dunno
            maybe_now = []
            if is_always_running(unit_file):
                maybe_now = ['--now']
            check_call(_systemctl('enable', unit_file, '--quiet', *maybe_now))
        elif unit.endswith('.timer'):
            check_call(_systemctl('enable', unit_file, '--now'))
        elif unit.endswith('.plist'):
            launchd.launchctl_load(unit_file=unit_file)
        else:
            raise AssertionError(a)

    # TODO not sure if this reload is even necessary??
    _daemon_reload()


def manage(state: State) -> None:
    apply_state(pending=state)


Error = str
# TODO perhaps, return Plan or error instead?


# eh, implicit convention that only one state will be emitted. oh well
# FIXME rename from lint? just use compileall or something as a syntax check?
def lint(tab_module: str) -> Iterator[Exception | State]:
    # TODO tbh compileall is pointless
    # - we can't find out source names property without importing
    # - we'll find out about errors during importing anyway

    try:
        jobs = load_jobs(tab_module)
    except Exception as e:
        # TODO could add better logging here? 'i.e. error while loading jobs'
        logger.exception(e)
        yield e
        return

    try:
        state = list(make_state(jobs))
    except Exception as e:
        logger.exception(e)
        yield e
        return

    yield state


def do_lint(tab_module: str) -> State:
    eit, vit = tee(lint(tab_module))
    errors = [r for r in eit if isinstance(r, Exception)]
    values = [r for r in vit if not isinstance(r, Exception)]
    assert len(errors) == 0, errors
    [state] = values
    return state


def _import_jobs(tab_module: str) -> list[Job]:
    module = importlib.import_module(tab_module)
    jobs_gen = getattr(module, 'jobs')  # get dynamically to make type checking happy
    return list(jobs_gen())


def load_jobs(tab_module: str) -> Iterator[Job]:
    # actually import in a separate process to avoid mess with polluting sys.modules
    # shouldn't be a problem in most cases, but it was annoying during tests
    with ProcessPoolExecutor(max_workers=1) as pool:
        jobs = pool.submit(_import_jobs, tab_module).result()

    emitted: dict[str, Job] = {}
    for job in jobs:
        assert isinstance(job, Job), job  # just in case for dumb typos
        assert job.unit_name not in emitted, (job, emitted[job.unit_name])
        yield job
        emitted[job.unit_name] = job


def apply(tab_module: str) -> None:
    # TODO rename do_lint to get_state?
    state = do_lint(tab_module)
    manage(state=state)


get_entries_for_monitor = systemd.get_entries_for_monitor if IS_SYSTEMD else launchd.get_entries_for_monitor


def main() -> None:
    from . import cli

    cli.main()


if __name__ == '__main__':
    main()


# TODO stuff I learnt:
# TODO  systemd-analyze --user unit-paths
# TODO blame!
#  systemd-analyze verify -- check syntax

# TODO would be nice to revert... via contextmanager?
# TODO assert that managed by dron
# TODO not sure what rollback should do w.r.t to
# TODO perhaps, only reenable changed ones? ugh. makes it trickier...

# TODO wonder if I remove timers, do they drop counts?
# TODO FIXME ok, for now, it's fine, but with more sophisticated timers might be a bit annoying

# TODO use python's literate types?


# TODO wow, that's quite annoying. so timer has to be separate file. oh well.

# TODO tui for confirming changes, show short diff?

# TODO actually for me, stuff like 'hourly' makes little sense; I usually space out in time..

# https://bugs.python.org/issue31528 eh, probably can't use configparser.. plaintext is good enough though.


# TODO later, implement logic for cleaning up old jobs


# TODO not sure if should do one by one or all at once?
# yeah, makes sense to do all at once...
# TODO warn about dirty state?


# TODO test with 'fake' systemd dir?

# TODO the assumption is that managed jobs are not changed manually, or changed in a way that doesn't break anything
# in general it's impossible to prevent anyway

# def update_unit(unit_file: Unit, old_body: Body, new_body: Body) -> Action:
#     if old_body == new_body:
#         pass # TODO no-op?
#     else:
#         raise RuntimeError(unit_file, old_body, new_body)
#     # TODO hmm FIXME!! yield is a nice way to make function lazy??


# TODO that perhaps? https://askubuntu.com/a/897317/427470



from __future__ import annotations

import sys
from collections.abc import Iterator
from pathlib import Path

import pytest

from ..dron import do_lint, load_jobs


@pytest.fixture
def tmp_pythonpath(tmp_path: Path) -> Iterator[Path]:
    ps = str(tmp_path)
    assert ps not in sys.path  # just in case
    sys.path.insert(0, ps)
    try:
        yield tmp_path
    finally:
        sys.path.remove(ps)


def test_load_jobs_basic(tmp_pythonpath: Path) -> None:
    tpath = Path(tmp_pythonpath) / 'test_drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job


def jobs() -> Iterator[Job]:
    job3 = job(
        '03:10',
        ['/path/to/command.py', 'some', 'args', '3'],
        unit_name='job3',
    )
    job1 = job(
        '01:10',
        ['/path/to/command.py', 'some', 'args', '1'],
        unit_name='job1',
    )
    yield job1
    yield job(
        '02:10',
        ['/path/to/command.py', 'some', 'args', '2'],
        unit_name='job2',
    )
    yield job3

'''
    )
    loaded = list(load_jobs(tab_module='test_drontab'))
    [job1, job2, job3] = loaded

    assert job1.when == '01:10'
    assert job1.command == ['/path/to/command.py', 'some', 'args', '1']
    assert job1.unit_name == 'job1'

    assert job2.when == '02:10'
    assert job2.command == ['/path/to/command.py', 'some', 'args', '2']
    assert job2.unit_name == 'job2'

    assert job3.when == '03:10'
    assert job3.command == ['/path/to/command.py', 'some', 'args', '3']
    assert job3.unit_name == 'job3'


def test_load_jobs_dupes(tmp_pythonpath: Path) -> None:
    tpath = Path(tmp_pythonpath) / 'test_drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job

def jobs() -> Iterator[Job]:
    yield job('00:00', 'echo', unit_name='job3')
    yield job('00:00', 'echo', unit_name='job1')
    # whoops! duplicate job name
    yield job('00:00', 'echo', unit_name='job3')
'''
    )
    with pytest.raises(AssertionError):
        _loaded = list(load_jobs(tab_module='test_drontab'))


def test_jobs_auto_naming(tmp_pythonpath: Path) -> None:
    tpath = Path(tmp_pythonpath) / 'test_drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job


job2 = job(
    '00:02',
    'echo',
)


def job_maker(when) -> Job:
    return job(when, 'echo job maker', stacklevel=2)


def jobs() -> Iterator[Job]:
    job_1 = job('00:01',
        'echo',
    )
    yield job2
    yield job('00:00', 'echo', unit_name='job_named')
    yield job_1
    job4 = \
       job('00:04', 'echo')
    job5     = job_maker('00:05')
    yield job5
    yield job4
'''
    )
    loaded = list(load_jobs(tab_module='test_drontab'))
    (job2, job_named, job_1, job5, job4) = loaded
    assert job_1.unit_name == 'job_1'
    assert job_1.when == '00:01'
    assert job2.unit_name == 'job2'
    assert job2.when == '00:02'
    assert job_named.unit_name == 'job_named'
    assert job_named.when == '00:00'
    assert job4.unit_name == 'job4'
    assert job4.when == '00:04'
    assert job5.unit_name == 'job5'
    assert job5.when == '00:05'


def test_do_lint(tmp_pythonpath: Path) -> None:
    def OK(body: str) -> None:
        tpath = Path(tmp_pythonpath) / 'test_drontab.py'
        tpath.write_text(body)
        do_lint(tab_module='test_drontab')

    def FAILS(body: str) -> None:
        with pytest.raises(Exception):
            OK(body)

    FAILS(
        body='''
    None.whatever
    '''
    )

    # no jobs
    FAILS(
        body='''
    '''
    )

    OK(
        body='''
def jobs():
    yield from []
'''
    )

    OK(
        body='''
from dron.api import job
def jobs():
    yield job(
        'hourly',
        ['/bin/echo', '123'],
        unit_name='unit_test',
    )
'''
    )

    from ..systemd import _is_missing_systemd

    if not _is_missing_systemd():
        from ..cli import _drontab_example

        # this test doesn't work without systemd yet, because launchd adapter doesn't support unquoted commands, at least yet..
        example = _drontab_example()
        # ugh. some hackery to make it find the executable..
        echo = " '/bin/echo"
        example = (
            example.replace(" 'linkchecker", echo)
            .replace(" '/home/user/scripts/run-borg", echo)
            .replace(" 'ping", " '/bin/ping")
        )
        OK(body=example)



from __future__ import annotations

import socket
import sys
from pprint import pprint

import click

from . import common, launchd, systemd
from .api import UnitName
from .common import (
    IS_SYSTEMD,
    MonitorParams,
    Unit,
    escape,
    logger,
    print_monitor,
    set_verify_off,
)
from .dron import (
    apply,
    do_lint,
    get_entries_for_monitor,
    load_jobs,
    manage,
    managed_units,
)


# TODO test it on CI?
# TODO explicitly inject it into readme?
def _drontab_example() -> str:
    return '''
from dron.api import job

# at the moment you're expected to define jobs() function that yields jobs
# in the future I might add more mechanisms
def jobs():
    # simple job that doesn't do much
    yield job(
        'daily',
        '/home/user/scripts/run-borg /home/user',
        unit_name='borg-backup-home',
    )

    yield job(
        'daily',
        'linkchecker https://beepb00p.xyz',
        unit_name='linkchecker-beepb00p',
    )

    # drontab is simply python code!
    # so if you're annoyed by having to rememver Systemd syntax, you can use a helper function
    def every(*, mins: int) -> str:
        return f'*:0/{mins}'

    # make sure my website is alive, it will send local email on failure
    yield job(
        every(mins=10),
        'ping https://beepb00p.xyz',
        unit_name='ping-beepb00p',
    )
'''.lstrip()


def _get_epilog() -> str:
    return '''
* Why?

In short, because I want to benefit from the heavy lifting that Systemd does: timeouts, resource management, restart policies, powerful scheduling specs and logging,
while not having to manually manipulate numerous unit files and restart the daemon all over.

I elaborate on what led me to implement it and motivation [[https://beepb00p.xyz/scheduler.html#what_do_i_want][here]]. Also:

\b
- why not just use [[https://beepb00p.xyz/scheduler.html#cron][cron]]?
- why not just use [[https://beepb00p.xyz/scheduler.html#systemd][systemd]]?
'''.strip()


@click.group(
    context_settings={'show_default': True},
    help="""
dron -- simple frontend for Systemd, inspired by cron.

\b
- *d* stands for 'Systemd'
- *ron* stands for 'cron'

dron is my attempt to overcome things that make working with Systemd tedious
""".strip(),
    epilog=_get_epilog(),
)
@click.option(
    '--marker',
    required=False,
    help=f'Use custom marker instead of default `{common.MANAGED_MARKER}`. Useful for developing/testing.',
)
def cli(*, marker: str | None) -> None:
    if marker is not None:
        common.MANAGED_MARKER = marker


arg_tab_module = click.option(
    '--module',
    'tab_module',
    type=str,
    default=f'drontab.{socket.gethostname()}',
)


# specify in readme???
# would be nice to use external checker..
# https://github.com/systemd/systemd/issues/8072
# https://unix.stackexchange.com/questions/493187/systemd-under-ubuntu-18-04-1-fails-with-failed-to-create-user-slice-serv
def _set_verify_off(ctx, param, value) -> None:  # noqa: ARG001
    if value is True:
        set_verify_off()


arg_no_verify = click.option(
    '--no-verify',
    is_flag=True,
    callback=_set_verify_off,
    expose_value=False,
    help='Skip systemctl verify step',
)


@cli.command('lint')
@arg_tab_module
@arg_no_verify
def cmd_lint(*, tab_module: str) -> None:
    # FIXME how to disable verity?
    # FIXME lint command isn't very interesting now btw?
    # perhaps instead, either add dry mode to apply
    # or split into the 'diff' part and side effect apply part
    _state = do_lint(tab_module)
    logger.info('all good')


@cli.command('print')
@arg_tab_module
@click.option('--pretty', is_flag=True, help='Pretty print')
@arg_no_verify
def cmd_print(*, tab_module: str, pretty: bool) -> None:
    """Parse and print drontab"""
    jobs = list(load_jobs(tab_module=tab_module))

    if pretty:
        import tabulate

        items = [
            {
                'UNIT': job.unit_name,
                'SCHEDULE': job.when,
                'COMMAND': escape(job.command),
            }
            for job in jobs
        ]
        print(tabulate.tabulate(items, headers="keys"))
    else:
        for j in jobs:
            print(j)


# TODO --force?
@cli.command('apply')
@arg_tab_module
def cmd_apply(*, tab_module: str) -> None:
    """Apply drontab (like 'crontab' with no args)"""
    apply(tab_module)


@cli.command('debug')
def cmd_debug() -> None:
    """Print some debug info"""
    managed = managed_units(with_body=False)  # TODO not sure about body
    for x in managed:
        pprint(x, stream=sys.stderr)


@cli.command('uninstall')
def cmd_uninstall() -> None:
    """Remove all managed jobs (will ask for confirmation)"""
    click.confirm('Going to remove all dron managed jobs. Continue?', default=True, abort=True)
    manage([])


@cli.group('job')
def cli_job() -> None:
    """Actions on individual jobs"""
    pass


def _prompt_for_unit() -> UnitName:
    from prompt_toolkit import PromptSession
    from prompt_toolkit.completion import WordCompleter

    # TODO print options
    managed = list(managed_units(with_body=False))
    units = [x.unit_file.stem for x in managed]

    print('Units under dron:', file=sys.stderr)
    for u in units:
        print(f'- {u}', file=sys.stderr)

    completer = WordCompleter(units, ignore_case=True)
    session = PromptSession("Select a unit: ", completer=completer)  # type: ignore[var-annotated]
    selected = session.prompt()
    return selected


arg_unit = click.argument('unit', type=Unit, default=_prompt_for_unit)


@cli_job.command('past')
@arg_unit
def cmd_past(unit: Unit) -> None:
    if IS_SYSTEMD:
        # TODO hmm seems like this just exit with 0 if unit diesn't exist
        return systemd.cmd_past(unit)
    else:
        return launchd.cmd_past(unit)


@cli_job.command('run')
@arg_unit
@click.option('--exec', 'do_exec', is_flag=True, help='Run directly, not via systemd/launchd')
def cmd_run(*, unit: Unit, do_exec: bool) -> None:
    """Run the job right now, ignoring the timer"""
    if IS_SYSTEMD:
        return systemd.cmd_run(unit=unit, do_exec=do_exec)
    else:
        return launchd.cmd_run(unit=unit, do_exec=do_exec)


@cli.command('monitor')
@click.option('-n', type=float, default=1.0, help='refresh every n seconds')
@click.option('--once', is_flag=True, help='only call once')
@click.option('--rate', is_flag=True, help='Display success rate (unstable and potentially slow)')
@click.option('--command', is_flag=True, help='Display command')
def cmd_monitor(*, n: float, once: bool, rate: bool, command: bool) -> None:
    """Monitor services/timers managed by dron"""
    params = MonitorParams(
        with_success_rate=rate,
        with_command=command,
    )

    if once:
        # old style monitor
        # TODO think if it's worth integrating with timers?
        managed = list(managed_units(with_body=False))  # body slows down this call quite a bit
        if len(managed) == 0:
            logger.warning('no managed units!')

        logger.debug('starting monitor...')

        entries = get_entries_for_monitor(managed=managed, params=params)
        print_monitor(entries)
    else:
        from .monitor import MonitorApp

        app = MonitorApp(monitor_params=params, refresh_every=n)
        app.run()


def main() -> None:
    cli()



import socket
from collections.abc import Iterator
from subprocess import PIPE, Popen

from .common import get_last_systemd_log, get_parser, get_stdin


def send_payload(payload: Iterator[bytes]) -> None:
    with Popen(['sendmail', '-t'], stdin=PIPE) as po:
        stdin = po.stdin
        assert stdin is not None
        for line in payload:
            stdin.write(line)
        stdin.flush()
    rc = po.poll()
    assert rc == 0, rc


def send_email(*, to: str, job: str, stdin: bool) -> None:
    def payload() -> Iterator[bytes]:
        hostname = socket.gethostname()
        yield f'''
To: {to}
From: dron <root@{hostname}>
Subject: {job}
Content-Transfer-Encoding: 8bit
Content-Type: text/plain; charset=UTF-8
'''.lstrip().encode('utf8')
        last_log = get_stdin() if stdin else get_last_systemd_log(job)
        yield from last_log

    send_payload(payload())


def main() -> None:
    p = get_parser()
    p.add_argument('--to', required=True)
    args = p.parse_args()
    send_email(to=args.to, job=args.job, stdin=args.stdin)


if __name__ == '__main__':
    main()



from .common import IS_SYSTEMD, get_parser
from .ntfy_common import run_ntfy

BACKEND = 'linux' if IS_SYSTEMD else 'darwin'


def main() -> None:
    p = get_parser()
    args = p.parse_args()

    run_ntfy(job=args.job, backend=BACKEND)


if __name__ == '__main__':
    main()



"""
uses https://github.com/dschep/ntfy
"""

import logging
import socket
import subprocess
import sys
from typing import NoReturn


# ty doesn't support NoReturn yet, see https://github.com/astral-sh/ty/issues/180
def run_ntfy(*, job: str, backend: str) -> NoReturn:  # ty: ignore[invalid-return-type]
    # TODO not sure what to do with --stdin arg here?
    # could probably use last N lines of log or something
    # TODO get last logs here?
    title = f'dron[{socket.gethostname()}]: {job} failed'
    body = title
    try:
        subprocess.check_call(['ntfy', '-b', backend, '-t', title, 'send', body])
    except Exception as e:
        logging.exception(e)  # noqa: LOG015
        # TODO fallback on email?
        sys.exit(1)
    sys.exit(0)



from __future__ import annotations

from dataclasses import asdict, fields
from functools import lru_cache
from typing import Any

from textual import events, work
from textual.app import App, ComposeResult
from textual.widgets import DataTable, Log

from .common import MonitorEntry, MonitorParams, unwrap
from .dron import get_entries_for_monitor, managed_units


# todo try RichLog? https://textual.textualize.io/guide/input
class LogHeader(Log):
    DEFAULT_CSS = """
    LogHeader {
        height: 5;
        dock: top;
    }
    """


MonitorEntries = dict[str, MonitorEntry]


@lru_cache(None)
def get_columns() -> list[str]:
    cols: list[str] = []
    for f in fields(MonitorEntry):
        cols.append(f.name)

    # TODO how to check it statically? MonitorEntry.pid isn't giving anything?
    cols.remove('pid')  # probs don't need?
    cols.remove('status_ok')
    return cols


def as_row(entry: MonitorEntry) -> dict[str, Any]:
    cols = get_columns()
    res = {k: v for k, v in asdict(entry).items() if k in cols}

    color = 'green' if entry.status_ok else 'red'
    res['status'] = f'[{color}]' + res['status'] + f'[/{color}]'

    # meh. workaround for next/left being max datetime
    if res['next'].startswith('9999-'):
        res['left'] = '--'
        res['next'] = '[yellow]' + 'never' + '[/yellow]'

    if entry.pid is not None:
        res['left'] = '--'
        res['next'] = '[yellow]' + 'running' + '[/yellow]'
    return res


class MonitorApp(App):
    def __init__(self, *, monitor_params: MonitorParams, refresh_every: float) -> None:
        super().__init__()
        self.monitor_params = monitor_params
        self.refresh_every = refresh_every

    def compose(self) -> ComposeResult:
        # TODO self.log is already defined?? what is it?
        self.logx = LogHeader(auto_scroll=True, highlight=True)

        # TODO input field to filter out jobs?

        # doesn't do anything??
        # self.log("HELLOOO")

        self.table: DataTable = DataTable(
            cursor_type='row',
            zebra_stripes=True,  # alternating colours
        )

        # NOTE: useful for debugging
        # yield self.logx
        yield self.table

    def on_mount(self) -> None:
        table = self.table

        for col in get_columns():
            table.add_column(label=col, key=col)

        # todo how to do async here as well?
        entries = self.get_entries()
        self.update(entries)

        self.set_focus(table)

        # TODO run and update it continuously? not sure
        # what if it isn't computed within interval?
        self.set_interval(interval=self.refresh_every, callback=self.update_in_background)

    def get_entries(self) -> MonitorEntries:
        managed = list(managed_units(with_body=False))  # body slows down this call quite a bit
        entries = get_entries_for_monitor(managed=managed, params=self.monitor_params)
        return {e.unit: e for e in entries}

    def update(self, entries: MonitorEntries) -> None:
        table = self.table

        # self.logx.write_line(f"HI {datetime.now().isoformat()}")

        current_rows: set[str] = {unwrap(x.value) for x in table.rows}
        to_remove = {x for x in current_rows if x not in entries}
        to_add = {x: entry for x, entry in entries.items() if x not in current_rows}
        to_update = {x: entry for x, entry in entries.items() if x in current_rows}

        for row_key in to_remove:
            table.remove_row(row_key=row_key)

        for row_key, entry in to_add.items():
            vals = as_row(entry).values()
            table.add_row(*vals, key=row_key)

        for row_key, entry in to_update.items():
            for col, value in as_row(entry).items():
                table.update_cell(row_key=row_key, column_key=col, value=value, update_width=True)

        columns = get_columns()

        def sort_key(row):
            data = dict(zip(columns, row, strict=True))
            is_running = 'running' in data['next']
            failed = 'exit-code' in data['status']
            return (not is_running, not failed, data['unit'])

        # TODO hmm kinda annoying, doesn't look like it preserves cursor position
        #  if the item pops on top of the list when a service is running?
        #  but I guess not a huge deal now
        table.sort(*columns, key=sort_key)

    @work(exclusive=True, thread=True)
    def update_in_background(self) -> None:
        entries = self.get_entries()
        self.call_from_thread(self.update, entries)

    def on_key(self, event: events.Key) -> None:
        actions = {
            'j': self.table.action_cursor_down,
            'k': self.table.action_cursor_up,
            'h': self.table.action_cursor_left,
            'l': self.table.action_cursor_right,
        }
        action = actions.get(event.key)
        if action is not None:
            action()
            return
        if event.key == 'q':
            self.exit()
        # TODO would be nice to display log on enter press or something?



from __future__ import annotations

import itertools
import json
import os
import re
import shlex
import sys
import textwrap
from collections.abc import Iterator, Sequence
from datetime import timedelta
from pathlib import Path
from subprocess import PIPE, Popen, check_call, check_output
from tempfile import TemporaryDirectory
from typing import Any

from .api import (
    OnCalendar,
    OnFailureAction,
    When,
)
from .common import (
    ALWAYS,
    MANAGED_MARKER,
    Command,
    LaunchdUnitState,
    MonitorEntry,
    MonitorParams,
    State,
    Unit,
    UnitFile,
    logger,
    unwrap,
)

# TODO custom launchd domain?? maybe instead could do dron/ or something?
_LAUNCHD_DOMAIN = f'gui/{os.getuid()}'


# in principle not necessary...
# but makes it much easier to filter out logs & lobs from launchctl dump
DRON_PREFIX = 'dron.'


def _launchctl(*args: Path | str) -> list[Path | str]:
    return ['launchctl', *args]


def _launch_agent(path: str) -> Path:
    # symlink for autostart
    assert path.endswith('.plist'), path  # meh
    assert not Path(path).is_absolute(), path

    LA = Path('~/Library/LaunchAgents').expanduser()
    link = LA / path
    return link


def fqn(name: Unit) -> str:
    return _LAUNCHD_DOMAIN + '/' + DRON_PREFIX + name


def launchctl_load(*, unit_file: UnitFile) -> None:
    # bootstrap is nicer than load
    # load is super defensive, returns code 0 on errors
    check_call(_launchctl('bootstrap', _LAUNCHD_DOMAIN, unit_file))
    _launch_agent(unit_file.name).symlink_to(unit_file)


def launchctl_unload(*, unit: Unit) -> None:
    # bootout is more verbose than unload
    # in addition unload is super defensive, returns code 0 on errors
    check_call(_launchctl('bootout', fqn(unit)))
    _launch_agent(unit + '.plist').unlink()


def launchctl_kickstart(*, unit: Unit) -> None:
    check_call(_launchctl('kickstart', fqn(unit)))


def launchctl_reload(*, unit: Unit, unit_file: UnitFile) -> None:
    # don't think there is a better way?
    launchctl_unload(unit=unit)
    launchctl_load(unit_file=unit_file)


def launchd_wrapper(*, job: str, on_failure: list[str]) -> list[str]:
    return [
        sys.executable,
        '-B',  # do not write byte code, otherwise it shits into dron directory if we're using editable install
        '-m', 'dron.launchd_wrapper',
        *itertools.chain.from_iterable(('--notify', n) for n in on_failure),
        '--job', job,
        '--',
    ]  # fmt: skip


def remove_launchd_wrapper(cmd: str) -> str:
    if ' dron.launchd_wrapper ' not in cmd:
        return cmd
    # uhh... not super reliable, but this is only used for monitor so hopefully fine
    [_, cmd] = cmd.split(' -- ', maxsplit=1)
    return cmd


def plist(
    *,
    unit_name: str,
    command: Command,
    on_failure: Sequence[OnFailureAction],
    when: When | None = None,
) -> str:
    # TODO hmm, kinda mirrors 'escape' method, not sure
    cmd: Sequence[str]
    if isinstance(command, (list, tuple)):
        cmd = tuple(map(str, command))
    elif isinstance(command, Path):
        cmd = [str(command)]
    elif isinstance(command, str) and ' ' not in command:
        cmd = [command]
    else:
        # unquoting and splitting is way trickier than quoting and joining...
        # not sure how to implement it p
        # maybe we just want bash -c in this case, dunno how to implement properly
        raise RuntimeError(command)
    del command

    mschedule = ''
    if when is None:
        # support later
        raise RuntimeError(unit_name)

    if when == ALWAYS:
        mschedule = '<key>KeepAlive</key>\n<true/>'
    else:
        assert isinstance(when, OnCalendar), when
        # https://www.freedesktop.org/software/systemd/man/systemd.time.html#
        # fmt: off
        seconds = {
            'minutely': 60,
            'hourly'  : 60 * 60,
            'daily'   : 60 * 60 * 24,
        }.get(when)
        # fmt: on
        if seconds is None:
            # ok, try systemd-like spec..
            # fmt: off
            specs = [
                (re.escape('*:0/')   + r'(\d+)', 60),
                (re.escape('*:*:0/') + r'(\d+)', 1),
            ]
            # fmt: on
            for rgx, mult in specs:
                m = re.fullmatch(rgx, when)
                if m is not None:
                    num = m.group(1)
                    seconds = int(num) * mult
                    break
        if seconds is None:
            # try to parse as hh:mm at least
            m = re.fullmatch(r'(\d\d):(\d\d)', when)
            assert m is not None, when
            hh = m.group(1)
            mm = m.group(2)
            mschedule = '\n'.join(
                [
                    '<key>StartCalendarInterval</key>',
                    '<dict>',
                    '<key>Hour</key>',
                    f'<integer>{int(hh)}</integer>',
                    '<key>Minute</key>',
                    f'<integer>{int(mm)}</integer>',
                    '</dict>',
                ]
            )
        else:
            mschedule = '\n'.join(('<key>StartInterval</key>', f'<integer>{seconds}</integer>'))

    assert mschedule != '', unit_name

    # meh.. not sure how to reconcile it better with systemd
    on_failure = [x.replace('--job %n', f'--job {unit_name}') + ' --stdin' for x in on_failure]

    # attempt to set argv[0] properly
    # hmm I was hoping it would make desktop notifications ('background service added' nicer)
    # but even after that it still only shows executable script name. ugh
    # program_argv = (unit_name, *cmd[1:])
    program_argv = (
        *launchd_wrapper(job=unit_name, on_failure=on_failure),
        *cmd,
    )
    del cmd
    program_argvs = '\n'.join(f'<string>{c}</string>' for c in program_argv)

    # TODO add log file, although mailer is already capturing stdout
    # TODO hmm maybe use the same log file for all dron jobs? would make it easier to rotate?
    res = f'''
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>

    <key>Label</key>
    <string>{DRON_PREFIX}{unit_name}</string>
    <key>ProgramArguments</key>
    <array>
{textwrap.indent(program_argvs, " " * 8)}
    </array>

    <key>RunAtLoad</key>
    <true/>

{textwrap.indent(mschedule, " " * 8)}

    <key>Comment</key>
    <string>{MANAGED_MARKER}</string>
</dict>
</plist>
'''.lstrip()
    return res


from .common import LaunchdUnitState


def launchd_state(*, with_body: bool) -> Iterator[LaunchdUnitState]:
    # sadly doesn't look like it has json interface??
    dump = check_output(['launchctl', 'dumpstate']).decode('utf8')

    name: str | None = None
    extras: dict[str, Any] = {}
    arguments: list[str] | None = None
    all_props: str | None = None
    fields = [
        'path',
        'last exit code',
        'pid',
        'run interval',
    ]
    for line in dump.splitlines():
        if name is None:
            # start of job description group
            name = line.removesuffix(' = {')
            all_props = ''
            continue
        elif line == '}':
            # end of job description group
            path: str | None = extras.get('path')
            if path is not None and 'dron' in path:
                # otherwsie likely some sort of system unit
                unit_file = Path(path)
                body = unit_file.read_text() if with_body else None

                # TODO extract 'state'??

                periodic_schedule = extras.get('run interval')
                calendal_schedule = 'com.apple.launchd.calendarinterval' in unwrap(all_props)

                schedule: str | None
                if periodic_schedule is not None:
                    schedule = 'every ' + periodic_schedule
                elif calendal_schedule:
                    # TODO parse properly
                    schedule = 'calendar'
                else:
                    # NOTE: seems like keepalive attribute isn't present in launchd dumpstate output
                    schedule = 'always'

                yield LaunchdUnitState(
                    unit_file=Path(path),
                    body=body,
                    cmdline=tuple(extras['arguments']),
                    # might not be present when we killed process manually?
                    last_exit_code=extras.get('last exit code'),
                    # pid might not be present (presumably when it's not running)
                    pid=extras.get('pid'),
                    schedule=schedule,
                )
            name = None
            all_props = None
            extras = {}
            continue

        all_props = unwrap(all_props) + line + '\n'

        if arguments is not None:
            if line == '\t}':
                extras['arguments'] = arguments
                arguments = None
            else:
                arg = line.removeprefix('\t\t')
                arguments.append(arg)
        else:
            xx = line.removeprefix('\t')
            for f in fields:
                zz = f'{f} = '
                if xx.startswith(zz):
                    extras[f] = xx.removeprefix(zz)
                    break
            # special handling..
            if xx.startswith('arguments = '):
                arguments = []


def verify_unit(*, unit_name: str, body: str) -> None:
    with TemporaryDirectory() as tdir:
        tfile = Path(tdir) / unit_name
        tfile.write_text(body)
        check_call(
            [
                'plutil',
                '-lint',
                '-s',  # silent on success
                tfile,
            ]
        )


def cmd_past(unit: Unit) -> None:
    sub = fqn('dron.' + unit)
    # fmt: off
    cmd = [
        # todo maybe use 'stream'??
        'log', 'show', '--info',
        # '--last', '24h',
        # hmm vvv that doesn't work, if we pass pid, predicate is ignored?
        # '--process', '1',
        # hmm, oddly enough "&&" massively slows the predicate??
        #'--predicate', f'processIdentifier=1 && (subsystem contains "gui/501/dron.{unit}")',
        '--predicate', f'subsystem contains "{sub}"',
        '--style', 'ndjson',
        '--color', 'always',
    ]
    # fmt: on
    with Popen(cmd, stdout=PIPE, encoding='utf8') as p:
        out = p.stdout
        assert out is not None
        for line in out:
            j = json.loads(line)
            if j.get('finished') == 1:
                # last event at the very end
                continue
            subsystem = j['subsystem']
            # sometimes subsystem contains pid at the end, need to chop it off
            # also that's wjy we can't use "subsystem = " predicate :(
            subsystem = subsystem.split(' ')[0]
            if sub != subsystem:
                continue
            msg = j['eventMessage']

            interesting = re.search(' spawned .* because', msg) or 'exited ' in msg
            if not interesting:
                continue
            ts = j['timestamp']
            print(ts, sub, msg)


def cmd_run(*, unit: Unit, do_exec: bool) -> None:
    if not do_exec:
        launchctl_kickstart(unit=unit)
        return

    states = []
    for s in launchd_state(with_body=False):
        if s.unit_file.stem == unit:
            states.append(s)
    [state] = states
    cmdline = state.cmdline
    assert cmdline is not None, unit

    ## cut off launchd wrapper
    sep_i = cmdline.index('--')
    cmdline = cmdline[sep_i + 1 :]
    ##

    cmds = ' '.join(map(shlex.quote, cmdline))
    logger.info(f'running: {cmds}')
    os.execvp(
        cmdline[0],
        list(cmdline),
    )


def get_entries_for_monitor(managed: State, *, params: MonitorParams) -> list[MonitorEntry]:
    # for now kinda copy pasted from systemd

    entries: list[MonitorEntry] = []
    for s in managed:
        assert isinstance(s, LaunchdUnitState), s

        unit_file = s.unit_file
        name = unit_file.name.removesuffix('.plist')

        is_seconds = re.fullmatch(r'every (\d+) seconds', s.schedule or '')
        if is_seconds is not None:
            delta = timedelta(seconds=int(is_seconds.group(1)))
            # meh, but works for now
            ss = f'every {delta}'
        else:
            ss = str(s.schedule)

        schedule = ss
        command = None
        if params.with_command:
            cmdline = s.cmdline
            assert cmdline is not None, name  # not None for launchd units
            command = ' '.join(map(shlex.quote, cmdline))
            command = remove_launchd_wrapper(command)

        status_ok = s.last_exit_code == '0'
        status = 'success' if status_ok else f'exitcode {s.last_exit_code}'

        pid = s.pid

        entries.append(
            MonitorEntry(
                unit=name,
                status=status,
                left='n/a',
                next='n/a',
                schedule=schedule,
                command=command,
                pid=pid,
                status_ok=status_ok,
            )
        )
    return entries



#!/usr/bin/env python3
import argparse
import os
import shlex
import sys
from collections.abc import Iterator
from pathlib import Path
from subprocess import PIPE, STDOUT, Popen
from typing import NoReturn

from loguru import logger

LOG_DIR = Path('~/Library/Logs/dron').expanduser()

# OSX/launchd is a piece of shit and doesn't seem possible to just set it globally everywhere?
# this works: launchctl setenv PYTHONPYCACHEPREFIX $PYTHONPYCACHEPREFIX
# however unclear how to set it in a way that it's running before all other agents
# allegedly possible to use global LaunchDaemon running as root, but doesn't seem possible to execute launchctl commands as other user from launchd plist??
PYCACHE_PATH = Path('~/.cache/pycache').expanduser()


# ty doesn't support NoReturn yet, see https://github.com/astral-sh/ty/issues/180
def main() -> NoReturn:  # ty: ignore[invalid-return-type]
    p = argparse.ArgumentParser()
    p.add_argument('--notify', action='append')
    p.add_argument('--job', required=True)
    # hmm, this doesn't work with keyword args??
    # p.add_argument('cmd', nargs=argparse.REMAINDER)
    args, rest = p.parse_known_args()

    assert rest[0] == '--', rest
    cmd = rest[1:]

    notify_cmds = [] if args.notify is None else args.notify
    job = args.job

    LOG_DIR.mkdir(parents=True, exist_ok=True)
    log_file = LOG_DIR / f'{job}.log'

    logger.add(log_file, rotation='100 MB')  # todo configurable? or rely on osx rotation?

    env = {**os.environ}
    if "PYTHONPYCACHEPREFIX" not in env:
        env["PYTHONPYCACHEPREFIX"] = str(PYCACHE_PATH)

    # hmm, a bit crap transforming everything to stdout? but not much we can do?
    captured_log = []
    try:
        with Popen(cmd, stdout=PIPE, stderr=STDOUT, env=env) as po:
            out = po.stdout
            assert out is not None
            for line in out:
                captured_log.append(line)
                sys.stdout.buffer.write(line)
        rc = po.poll()

        if rc == 0:
            # short circuit
            sys.exit(0)
    except Exception as e:
        # Popen istelf still fail due to permission denied or something
        logger.exception(e)
        captured_log.append(str(e).encode('utf8'))
        rc = 123

    def payload() -> Iterator[bytes]:
        yield f"exit code: {rc}\n".encode()
        yield b'command: \n'
        yield (' '.join(map(shlex.quote, cmd)) + '\n').encode('utf8')
        yield f'log file: {log_file}\n'.encode()
        yield b'\n'
        yield b'output (stdout + stderr):\n\n'
        # TODO shit -- if multiple notifications, can't use generator for captured_log
        # unless we notify simultaneously?
        yield from captured_log

    for line in payload():
        logger.info(line.decode('utf8').rstrip('\n'))  # meh

    for notify_cmd in notify_cmds:
        logger.info(f'notifying: {notify_cmd}')
        try:
            with Popen(notify_cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE) as po:
                sin = po.stdin
                assert sin is not None
                for line in payload():
                    sin.write(line)
                (sout, serr) = po.communicate()
                for l in sout.decode('utf8').splitlines():
                    logger.debug(l)
                for l in serr.decode('utf8').splitlines():
                    logger.debug(l)
            assert po.poll() == 0, notify_cmd
        except Exception as e:
            logger.error(f'notificaiton failed: {notify_cmd}')
            logger.exception(e)

    sys.exit(rc)


if __name__ == '__main__':
    main()



"""
uses telegram-send for Telegram notifications
make sure to run "telegram-send --configure" beforehand!
"""

import asyncio
import logging
import socket
import sys

from .common import get_last_systemd_log, get_parser, get_stdin


def send(*, message: str) -> None:
    import telegram_send  # type: ignore[import-untyped]

    asyncio.run(telegram_send.send(messages=[message]))


def main() -> None:
    p = get_parser()
    args = p.parse_args()

    job: str = args.job
    stdin: bool = args.stdin

    body = f'dron[{socket.gethostname()}]: {job} failed'

    last_log = get_stdin() if stdin else get_last_systemd_log(job)
    body += '\n' + '\n'.join(l.decode('utf8') for l in last_log)

    try:
        send(message=body)
    except Exception as e:
        logging.exception(e)  # noqa: LOG015
        # TODO fallback on email?
        sys.exit(1)
    sys.exit(0)


if __name__ == '__main__':
    main()



from __future__ import annotations

import platform
import shlex
import sys
from collections.abc import Iterable, Sequence
from dataclasses import asdict, dataclass, replace
from datetime import datetime
from pathlib import Path
from typing import Any

from loguru import logger  # noqa: F401

datetime_aware = datetime
datetime_naive = datetime


# TODO can remove this? although might be useful for tests
VERIFY_UNITS: bool = True
# TODO ugh. verify tries using already installed unit files so if they were bad, everything would fail
# I guess could do two stages, i.e. units first, then timers
# dunno, a bit less atomic though...


def set_verify_off() -> None:
    global VERIFY_UNITS
    VERIFY_UNITS = False


@dataclass
class MonitorParams:
    with_success_rate: bool
    with_command: bool


Unit = str
Body = str
UnitFile = Path


@dataclass
class UnitState:
    unit_file: UnitFile
    body: Body | None
    cmdline: Sequence[str] | None  # can be None for timers


@dataclass
class SystemdUnitState(UnitState):
    dbus_properties: Any  # seems like keeping this around massively speeds up dbus access...


@dataclass
class LaunchdUnitState(UnitState):
    # NOTE: can legit be str (e.g. if unit was never ran before)
    last_exit_code: str | None
    pid: str | None
    schedule: str | None


State = Iterable[UnitState]


IS_SYSTEMD = platform.system() != 'Darwin'  # if not systemd it's launchd


def unwrap[T](x: T | None) -> T:
    assert x is not None
    return x


PathIsh = str | Path

# if it's an str, assume it's already escaped
# otherwise we are responsible for escaping..
Command = PathIsh | Sequence[PathIsh]


OnCalendar = str
TimerSpec = dict[str, str]  # meh # TODO why is it a dict???
ALWAYS = 'always'
When = OnCalendar | TimerSpec


MANAGED_MARKER: str = '(MANAGED BY DRON)'


def is_managed(body: str) -> bool:
    # switching off it because it's unfriendly to launchd
    legacy_marker = '<MANAGED BY DRON>'
    return MANAGED_MARKER in body or legacy_marker in body


pytest_fixture: Any
under_pytest = 'pytest' in sys.modules
if under_pytest:
    import pytest

    pytest_fixture = pytest.fixture
else:
    pytest_fixture = lambda f: f  # no-op otherwise to prevent pytest import


Escaped = str


def escape(command: Command) -> Escaped:
    if isinstance(command, Escaped):
        return command
    elif isinstance(command, Path):
        return escape([command])
    else:
        return ' '.join(shlex.quote(str(part)) for part in command)


def wrap(script: PathIsh, command: Command) -> Escaped:
    return shlex.quote(str(script)) + ' ' + escape(command)


def test_wrap() -> None:
    assert wrap('/bin/bash', ['-c', 'echo whatever']) == "/bin/bash -c 'echo whatever'"
    bin_ = Path('/bin/bash')
    assert wrap(bin_, "-c 'echo whatever'") == "/bin/bash -c 'echo whatever'"
    assert wrap(bin_, ['echo', bin_]) == "/bin/bash echo /bin/bash"
    assert wrap('cat', bin_) == "cat /bin/bash"


@dataclass(order=True)
class MonitorEntry:
    unit: str
    status: str
    left: str
    next: str
    schedule: str
    command: str | None
    pid: str | None

    """
    'status' is coming from systemd/launchd, and it's a string.

    So status_ok should be used instead if you actually want to rely on something robust.
    """
    status_ok: bool


def print_monitor(entries: Iterable[MonitorEntry]) -> None:
    entries = sorted(
        entries,
        key=lambda e: (e.pid is None, e.status_ok, e),
    )

    import tabulate
    import termcolor

    tabulate.PRESERVE_WHITESPACE = True

    headers = [
        'UNIT',
        'STATUS',
        'LEFT',
        'NEXT',
        'SCHEDULE',
    ]
    with_command = any(x.command is not None for x in entries)
    if with_command:
        headers.append('COMMAND')

    items = []
    for e in entries:
        e = replace(
            e,
            status=termcolor.colored(e.status, 'green' if e.status_ok else 'red'),
        )
        if e.pid is not None:
            e = replace(
                e,
                next=termcolor.colored('running', 'yellow'),
                left='--',
            )
        items.append(list(asdict(e).values())[: len(headers)])
    print(tabulate.tabulate(items, headers=headers))



from .common import get_parser
from .ntfy_common import run_ntfy


def main() -> None:
    p = get_parser()
    args = p.parse_args()
    run_ntfy(job=args.job, backend='telegram')


if __name__ == '__main__':
    main()



import argparse
import platform
import shlex
import sys
from collections.abc import Iterator
from subprocess import PIPE, STDOUT, Popen, check_output

IS_SYSTEMD = platform.system() != 'Darwin'  # if not systemd it's launchd


def get_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser()
    p.add_argument('--job', required=True)
    p.add_argument('--stdin', action='store_true')
    return p


def get_stdin() -> Iterator[bytes]:
    yield from sys.stdin.buffer


def get_last_systemd_log(job: str) -> Iterator[bytes]:
    # output unit status
    cmd = ['systemctl', '--user', 'status', '--no-pager', job, '-o', 'cat']
    yield b'$ ' + ' '.join(map(shlex.quote, cmd)).encode('utf8') + b'\n\n'
    with Popen(cmd, stdout=PIPE, stderr=STDOUT) as po:
        out = po.stdout
        assert out is not None
        yield from out
    rc = po.poll()
    assert rc in {
        0,
        3,  # 3 means failure due to job exit code
    }, rc

    # for logs, we used to use --lines 1000000 in systemctl status
    # however, from around 2024 it stated consuming too much time
    # (as if it actually retrieved 1000000 lines and only then tooks the ones relevant to the unit??)

    cmd = ['systemctl', '--user', 'show', job, '-p', 'InvocationID', '--value']
    invocation_id = check_output(cmd, text=True)
    invocation_id = invocation_id.strip()  # for some reason dumps multiple lines?
    assert len(invocation_id) > 0  # just in case, todo maybe make defensive?

    yield b'\n'
    cmd = ['journalctl', '--no-pager', f'_SYSTEMD_INVOCATION_ID={invocation_id}']
    yield b'$ ' + ' '.join(map(shlex.quote, cmd)).encode('utf8') + b'\n\n'
    with Popen(cmd, stdout=PIPE, stderr=STDOUT) as po:
        out = po.stdout
        assert out is not None
        yield from out
    rc = po.poll()
    assert rc == 0



import pytest


@pytest.fixture(scope='session', autouse=True)
def disable_verify_units_if_no_systemd():
    '''
    If we can't use systemd, we need to suppress systemd-specific linting
    '''
    from . import common
    from .systemd import _is_missing_systemd

    reason = _is_missing_systemd()
    if reason is not None:
        common.VERIFY_UNITS = False
    try:
        yield
    finally:
        common.VERIFY_UNITS = True



from __future__ import annotations

import json
import os
import re
import shlex
import shutil
from collections.abc import Iterator, Sequence
from datetime import UTC, datetime, timedelta
from functools import lru_cache
from itertools import groupby
from pathlib import Path
from subprocess import PIPE, Popen, run
from tempfile import TemporaryDirectory
from typing import Any
from zoneinfo import ZoneInfo

from .api import (
    OnFailureAction,
    When,
)
from .common import (
    MANAGED_MARKER,
    Body,
    Command,
    MonitorEntry,
    MonitorParams,
    State,
    SystemdUnitState,
    TimerSpec,
    Unit,
    datetime_aware,
    escape,
    is_managed,
    logger,
)


def _is_missing_systemd() -> str | None:
    has_systemd = shutil.which('systemctl') is not None
    if not has_systemd:
        return "systemd not available, running under docker or osx"
    return None


def _systemctl(*args: Path | str) -> list[Path | str]:
    return ['systemctl', '--user', *args]


def managed_header() -> str:
    return f'''
# {MANAGED_MARKER}
# If you do any manual changes, they will be overridden on the next dron run
'''.lstrip()


# TODO how to come up with good implicit job name?
# TODO do we need a special target for dron?
def timer(*, unit_name: str, when: When) -> str:
    spec: TimerSpec
    if isinstance(when, str):
        spec = {'OnCalendar': when}
    else:
        spec = when

    specs = '\n'.join(f'{k}={v}' for k, v in spec.items())

    return f'''
{managed_header()}
[Unit]
Description=Timer for {unit_name} {MANAGED_MARKER}

[Timer]
{specs}

[Install]
WantedBy=timers.target
'''.lstrip()


# TODO add Restart=always and RestartSec?
# TODO allow to pass extra args
def service(
    *,
    unit_name: str,
    command: Command,
    on_failure: Sequence[OnFailureAction],
    **kwargs: str,
) -> str:
    # TODO not sure if something else needs to be escaped for ExecStart??
    # todo systemd-escape? but only can be used for names

    # ok OnFailure is quite annoying since it can't take arguments etc... seems much easier to use ExecStopPost
    # (+ can possibly run on success too that way?)
    # https://unix.stackexchange.com/a/441662/180307
    cmd = escape(command)

    exec_stop_posts = [
        f"ExecStopPost=/bin/sh -c 'if [ $$EXIT_STATUS != 0 ]; then {action}; fi'" for action in on_failure
    ]

    sections: dict[str, list[str]] = {}
    sections['[Unit]'] = [f'Description=Service for {unit_name} {MANAGED_MARKER}']

    sections['[Service]'] = [
        f'ExecStart={cmd}',
        *exec_stop_posts,
    ]

    for k, value in kwargs.items():
        # ideally it would have section name
        m = re.search(r'(\[\w+\])(.*)', k)
        if m is not None:
            section = m.group(1)
            key = m.group(2)
        else:
            # 'legacy' behaviour, by default put into [Service]
            section = '[Service]'
            key = k
        if section not in sections:
            sections[section] = []
        sections[section].append(f'{key}={value}')

    res = managed_header()
    for section_name, lines in sections.items():
        res += '\n\n' + '\n'.join([section_name, *lines])
    res += '\n'

    return res


def test_managed() -> None:
    skip_if_no_systemd()
    from .dron import verify_unit

    assert is_managed(timer(unit_name='whatever', when='daily'))

    custom = '''
[Service]
ExecStart=/bin/echo 123
'''
    verify_unit(unit_name='other.service', body=custom)  # precondition
    assert not is_managed(custom)


def verify_units(pre_units: list[tuple[Unit, Body]]) -> None:
    # ugh. systemd-analyze takes about 0.2 seconds for each unit for some reason
    # oddly enough, in bulk it works just as fast :thinking_face:
    # also doesn't work in parallel (i.e. parallel processes)
    # that ends up with some weird errors trying to connect to socket
    with TemporaryDirectory() as _tdir:
        tdir = Path(_tdir)
        for unit, body in pre_units:
            (tdir / unit).write_text(body)
        res = run(['systemd-analyze', '--user', 'verify', *tdir.glob('*')], capture_output=True, check=False)
        # ugh. apparently even exit code 0 doesn't guarantee correct output??
        out = res.stdout.decode('utf8')
        err = res.stderr.decode('utf8')
        assert out == '', out
        if err == '':
            return

        err_lines = err.splitlines(keepends=True)
        unique_err_lines = []
        # uhh.. in bulk mode it spams with tons of 'Cannot add dependency job' for some reason
        # I guess it kinda treats everything as dependent on each other??
        # https://github.com/systemd/systemd/blob/b692ad36b99909453cf4f975a346e41d6afc68a0/src/core/transaction.c#L978
        for l in err_lines:
            if l not in unique_err_lines:
                unique_err_lines.append(l)
        err_lines = unique_err_lines

        if len(err_lines) == 0:
            return

        msg = f'failed checking , exit code {res.returncode}'
        logger.error(msg)
        logger.error('systemd-analyze output:')
        for line in err_lines:
            logger.error(line.strip())
        raise RuntimeError(msg)


def test_verify_systemd() -> None:
    skip_if_no_systemd()
    from .dron import verify_unit

    def FAILS(body: str) -> None:
        import pytest

        with pytest.raises(Exception):
            verify_unit(unit_name='whatever.service', body=body)

    def OK(body: str) -> None:
        verify_unit(unit_name='ok.service', body=body)

    OK(
        body='''
[Service]
ExecStart=/bin/echo 123
'''
    )

    from .api import notify

    on_failure = (
        notify.email('test@gmail.com'),
        notify.desktop_notification,
    )
    OK(body=service(unit_name='alala', command='/bin/echo 123', on_failure=on_failure))

    # garbage
    FAILS(body='fewfewf')

    # no execstart
    FAILS(
        body='''
[Service]
StandardOutput=journal
'''
    )

    FAILS(
        body='''
[Service]
ExecStart=yes
StandardOutput=baaad
'''
    )


def _sd(s: str) -> str:
    return f'org.freedesktop.systemd1{s}'


class BusManager:
    def __init__(self) -> None:
        # unused-ignore because on macos there is no dbus (but this code is still running mypy on CI)
        from dbus import (  # type: ignore[import-untyped,import-not-found,unused-ignore]
            Interface,
            SessionBus,
        )

        self.Interface = Interface  # meh

        self.bus = SessionBus()  # note: SystemBus is for system-wide services
        systemd = self.bus.get_object(_sd(''), '/org/freedesktop/systemd1')
        self.manager = Interface(systemd, dbus_interface=_sd('.Manager'))

    def properties(self, u: Unit):
        service_unit = self.manager.GetUnit(u)
        service_proxy = self.bus.get_object(_sd(''), str(service_unit))
        return self.Interface(service_proxy, dbus_interface='org.freedesktop.DBus.Properties')

    @staticmethod  # meh
    def prop(obj, schema: str, name: str):
        return obj.Get(_sd(schema), name)

    @classmethod
    def exec_start(cls, props) -> Sequence[str]:
        dbus_exec_start = cls.prop(props, '.Service', 'ExecStart')
        return [str(x) for x in dbus_exec_start[0][1]]


def systemd_state(*, with_body: bool) -> State:
    bus = BusManager()
    states = bus.manager.ListUnits()  # ok nice, it's basically instant

    for state in states:
        name = state[0]
        descr = state[1]
        if not is_managed(descr):
            continue

        # todo annoying, this call still takes some time... but whatever ok
        props = bus.properties(name)

        # useful for debugging, can also use .Service if it's not a timer
        # all_properties = props.GetAll(_sd('.Unit'))
        # however GetAll seems slower than gettind individual properties

        # stale = int(bus.prop(props, '.Unit', 'NeedDaemonReload')) == 1
        # TODO do we actually need to resolve?
        unit_file = Path(str(bus.prop(props, '.Unit', 'FragmentPath'))).resolve()
        body = unit_file.read_text() if with_body else None
        cmdline: Sequence[str] | None
        if '.timer' in name:  # meh
            cmdline = None
        else:
            cmdline = BusManager.exec_start(props)

        yield SystemdUnitState(unit_file=unit_file, body=body, cmdline=cmdline, dbus_properties=props)


def test_managed_units() -> None:
    skip_if_no_systemd()
    # TODO wonder if i'd be able to use launchd on ci...
    from .cli import cmd_monitor
    from .dron import managed_units

    # shouldn't fail at least
    list(managed_units(with_body=True))

    # TODO ugh. doesn't work on circleci, fails with
    # dbus.exceptions.DBusException: org.freedesktop.DBus.Error.BadAddress: Address does not contain a colon
    # todo maybe don't need it anymore with 20.04 circleci?
    if 'CI' not in os.environ:
        cmd_monitor.callback(n=1, once=True, command=True, rate=True)  # type: ignore[misc]


def skip_if_no_systemd() -> None:
    import pytest

    reason = _is_missing_systemd()
    if reason is not None:
        pytest.skip(f'No systemd: {reason}')


_UTCMAX = datetime.max.replace(tzinfo=UTC)


class MonitorHelper:
    def from_usec(self, usec) -> datetime_aware:
        u = int(usec)
        if u == 2**64 - 1:  # apparently systemd uses max uint64
            # happens if the job is running ATM?
            return _UTCMAX
        else:
            return datetime.fromtimestamp(u / 10**6, tz=UTC)

    @property
    @lru_cache  # noqa: B019
    def local_tz(self) -> ZoneInfo:
        try:
            # it's a required dependency, but still might fail in some weird environments?
            #   e.g. if zoneinfo information isn't available
            from tzlocal import get_localzone

            return get_localzone()
        except Exception:
            logger.error("Couldn't determine local timezone! Falling back to UTC")
            return ZoneInfo('UTC')


# TODO maybe format seconds prettier. dunno
def _fmt_delta(d: timedelta) -> str:
    # format to reduce constant countdown...
    ad = abs(d)
    # get rid of microseconds
    ad = ad - timedelta(microseconds=ad.microseconds)

    day = timedelta(days=1)
    hour = timedelta(hours=1)
    minute = timedelta(minutes=1)
    gt = False
    if ad > day:
        full_days = ad // day
        hours = (ad % day) // hour
        ads = f'{full_days}d {hours}h'
        gt = True
    elif ad > minute:
        full_mins = ad // minute
        ad = timedelta(minutes=full_mins)
        ads = str(ad)
        gt = True
    else:
        # show exact
        ads = str(ad)
    if len(ads) == 7:
        ads = '0' + ads  # meh. fix missing leading zero in hours..
    ads = ('>' if gt else '') + ads
    return ads


def get_entries_for_monitor(managed: State, *, params: MonitorParams) -> list[MonitorEntry]:
    mon = MonitorHelper()

    UTCNOW = datetime.now(tz=UTC)

    bus = BusManager()

    entries: list[MonitorEntry] = []

    # sort so that neigbouring unit.service/unit.timer go one after another for grouping
    sort_key = lambda unit: unit.unit_file.name
    # for grouping, group by common stem of timers and services
    stem_name = lambda unit: sort_key(unit).split('.')[0]
    for k, _gr in groupby(sorted(managed, key=sort_key), key=stem_name):
        gr: list[SystemdUnitState] = []
        for x in _gr:
            assert isinstance(x, SystemdUnitState), x  # guaranteed by managed_units function
            gr.append(x)

        # if timer is None, guess that means the job is always running?
        timer: SystemdUnitState | None
        service: SystemdUnitState
        if len(gr) == 2:
            [service, timer] = gr
        else:
            assert len(gr) == 1, gr
            [service] = gr
            timer = None

        service_props = service.dbus_properties

        if timer is not None:
            props = timer.dbus_properties
            # FIXME this might be io bound? maybe make async or use thread pool?
            cal = bus.prop(props, '.Timer', 'TimersCalendar')
            next_ = bus.prop(props, '.Timer', 'NextElapseUSecRealtime')

            # note: there is also bus.prop(props, '.Timer', 'LastTriggerUSec'), but makes more sense to use unit to account for manual runs
            last = bus.prop(service_props, '.Unit', 'ActiveExitTimestamp')

            schedule = cal[0][1]  # TODO is there a more reliable way to retrieve it??
            # todo not sure if last is really that useful..

            last_dt = mon.from_usec(last)
            next_dt = mon.from_usec(next_)
            nexts = next_dt.astimezone(mon.local_tz).replace(tzinfo=None, microsecond=0).isoformat()

            if next_dt == datetime.max:
                left_delta = timedelta(0)
            else:
                left_delta = next_dt - UTCNOW
        else:
            left_delta = timedelta(0)  # TODO
            last_dt = UTCNOW
            nexts = 'n/a'
            schedule = 'always'

        left = f'{_fmt_delta(left_delta)!s:<9}'
        if last_dt.timestamp() == 0:
            ago = 'never'  # TODO yellow?
        else:
            passed_delta = UTCNOW - last_dt
            ago = str(_fmt_delta(passed_delta))
        # TODO instead of hacking microsecond, use 'NOW' or something?

        # TODO some summary too? e.g. how often in failed
        if params.with_command:
            exec_start = service.cmdline
            assert exec_start is not None, service  # not None for services
            command = shlex.join(exec_start)
        else:
            command = None
        _pid: int | None = int(bus.prop(service_props, '.Service', 'MainPID'))
        pid = None if _pid == 0 else str(_pid)

        if params.with_success_rate:
            rate = _unit_success_rate(service.unit_file.name)
            rates = f' {rate:.2f}'
        else:
            rates = ''

        service_result = bus.prop(service_props, '.Service', 'Result')
        status_ok = service_result == 'success'
        status = f'{service_result:<9} {ago:<8}{rates}'

        entries.append(
            MonitorEntry(
                unit=k,
                status=status,
                left=left,
                next=nexts,
                schedule=schedule,
                command=command,
                pid=pid,
                status_ok=status_ok,
            )
        )
    return entries


Json = dict[str, Any]


def _unit_logs(unit: Unit) -> Iterator[Json]:
    # TODO so do I need to parse logs to get failure stats? perhaps json would be more reliable
    cmd = f'journalctl --user -u {unit} -o json -t systemd --output-fields UNIT_RESULT,JOB_TYPE,MESSAGE'
    with Popen(cmd.split(), stdout=PIPE) as po:
        stdout = po.stdout
        assert stdout is not None
        for line in stdout:
            j = json.loads(line.decode('utf8'))
            # apparently, successful runs aren't getting logged? not sure why
            # jt = j.get('JOB_TYPE')
            # ur = j.get('UNIT_RESULT')
            # not sure about this..
            yield j


def _unit_success_rate(unit: Unit) -> float:
    started = 0
    failed = 0
    # TODO not sure how much time it takes to query all journals?
    for j in _unit_logs(unit):
        jt = j.get('JOB_TYPE')
        ur = j.get('UNIT_RESULT')
        if jt is not None:
            assert ur is None
            started += 1
        elif ur is not None:
            assert jt is None
            failed += 1
        else:
            # TODO eh? sometimes jobs also report Succeeded status
            # e.g. syncthing-paranoid
            pass
    if started == 0:
        assert failed == 0, unit
        return 1.0
    success = started - failed
    return success / started


def cmd_past(unit: Unit) -> None:
    mon = MonitorHelper()
    for j in _unit_logs(unit):
        ts = mon.from_usec(j['__REALTIME_TIMESTAMP'])
        msg = j['MESSAGE']
        print(ts.isoformat(), msg)


def cmd_run(*, unit: Unit, do_exec: bool) -> None:
    assert do_exec  # support without exec later
    # TODO we might have called it before via managed_units.. maybe need to cache
    states = []
    for s in systemd_state(with_body=False):
        # meh
        unit_name = s.unit_file.name
        if unit_name.endswith('.timer'):
            continue
        if s.unit_file.stem == unit:
            states.append(s)
    [state] = states
    cmdline = state.cmdline
    assert cmdline is not None
    cmds = ' '.join(map(shlex.quote, cmdline))
    logger.info(f'running: {cmds}')
    os.execvp(
        cmdline[0],
        list(cmdline),
    )


# used to use this, keeping for now just for the refernce
# def old_systemd_emailer() -> None:
#     user = getpass.getuser()
#     X = textwrap.dedent(f'''
#     [Unit]
#     Description=status email for %i to {user}
#
#     [Service]
#     Type=oneshot
#     ExecStart={SYSTEMD_EMAIL} --to {user} --unit %i --journalctl-args "-o cat"
#     # TODO why these were suggested??
#     # User=nobody
#     # Group=systemd-journal
#     ''')
#
#     write_unit(unit=f'status-email@.service', body=X, prefix=SYSTEMD_USER_DIR)
#     # I guess makes sense to reload here; fairly atomic step
#     _daemon_reload()
